/***
 * Neuroph  http://neuroph.sourceforge.net
 * Copyright by Neuroph Project (C) 2008 
 *
 * This file is part of Neuroph framework.
 *
 * Neuroph is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Neuroph is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Neuroph. If not, see <http://www.gnu.org/licenses/>.
 */

package org.neuroph.util;

import java.util.Vector;

import org.neuroph.core.Layer;
import org.neuroph.core.NeuralNetwork;
import org.neuroph.core.Neuron;
import org.neuroph.nnet.Adaline;
import org.neuroph.nnet.CompetitiveNetwork;
import org.neuroph.nnet.SupervisedHebbianNetwork;
import org.neuroph.nnet.UnsupervisedHebbianNetwork;
import org.neuroph.nnet.Hopfield;
import org.neuroph.nnet.Kohonen;
import org.neuroph.nnet.MaxNet;
import org.neuroph.nnet.MultiLayerPerceptron;
import org.neuroph.nnet.Perceptron;
import org.neuroph.nnet.RbfNetwork;

/**
 * This class provides methods to create various neural networks.
 * @author Zoran Sevarac <sevarac@gmail.com>
 */
public class NeuralNetworkFactory {

	/**
	 * Creates  and returns an instance of Adaline network
	 * @param inputsNum
	 * @return
	 */
	public static Adaline createAdaline(int inputsNum) {
		Adaline nnet = new Adaline(new Integer(inputsNum));
		return nnet;
	}


	/**
	 * Creates  and returns an instance of Perceptron network
	 * @param inputsNum
	 * @return
	 */	
	public static Perceptron createPerceptron(int inputNeuronsNum, int outputNeuronsNum, TransferFunctionType transferFunctionType) {
		Perceptron nnet = new Perceptron(inputNeuronsNum, outputNeuronsNum, transferFunctionType);
		return nnet;
	}

	/**
	 * Creates and returns an instance of Multi Layer Perceptron
	 * @param layersStr
	 * @param transferFunctionType
	 * @return
	 */
	public static MultiLayerPerceptron createMLPerceptron(String layersStr, TransferFunctionType transferFunctionType) {
		Vector<Integer> layerSizes = VectorParser.parseInteger(layersStr);
		MultiLayerPerceptron nnet = new MultiLayerPerceptron(layerSizes,
				transferFunctionType);
		return nnet;
	}
	
	/**
	 * Creates and returns an instance of Hopfield network
	 * @param neuronsNum
	 * @return
	 */
	public static Hopfield createHopfield(int neuronsNum) {
		Hopfield nnet = new Hopfield(neuronsNum);
		return nnet;
	}

	/**
	 * Creates and returns an instance of Kohonen network
	 * @param in
	 * @param out
	 * @return
	 */
	public static Kohonen createKohonen(int in, int out) {
		Kohonen nnet = new Kohonen(new Integer(in), new Integer(out));
		return nnet;
	}

	/**
	 * Creates and returns an instance of Hebbian network
	 * @param inputNeuronsNum
	 * @param outputNeuronsNum
	 * @param transferFunctionType
	 * @return
	 */
	public static SupervisedHebbianNetwork createSupervisedHebbian(int inputNeuronsNum,
			int outputNeuronsNum, TransferFunctionType transferFunctionType) {
		SupervisedHebbianNetwork nnet = new SupervisedHebbianNetwork(inputNeuronsNum,
				outputNeuronsNum, transferFunctionType);
		return nnet;
	}
	
	/**
	 * Creates and returns an instance of Unsupervised Hebbian Network
	 * @param inputNeuronsNum
	 * @param outputNeuronsNum
	 * @param transferFunctionType
	 * @return
	 */
	public static UnsupervisedHebbianNetwork createUnsupervisedHebbian(int inputNeuronsNum,
			int outputNeuronsNum, TransferFunctionType transferFunctionType) {
		UnsupervisedHebbianNetwork nnet = new UnsupervisedHebbianNetwork(inputNeuronsNum,
				outputNeuronsNum, transferFunctionType);
		return nnet;
	}	

	/**
	 * Creates and returns an instance of Max Net network
	 * @param neuronsNum
	 * @return
	 */
	public static MaxNet createMaxNet(int neuronsNum) {
		MaxNet nnet = new MaxNet(neuronsNum);
		return nnet;
	}

	/**
	 * Creates and returns an instance of competitive network
	 * @param inputNeuronsNum
	 * @param outputNeuronsNum
	 * @return
	 */
	public static CompetitiveNetwork createCompetitiveNetwork(
			int inputNeuronsNum, int outputNeuronsNum) {
		CompetitiveNetwork nnet = new CompetitiveNetwork(inputNeuronsNum,
				outputNeuronsNum);
		return nnet;
	}

	/**
	 * Creates and returns an instance of RBF network
	 * @param inputNeuronsNum
	 * @param rbfNeuronsNum
	 * @param outputNeuronsNum
	 * @return
	 */
	public static RbfNetwork createRbfNetwork(int inputNeuronsNum,
			int rbfNeuronsNum, int outputNeuronsNum) {
		RbfNetwork nnet = new RbfNetwork(inputNeuronsNum, rbfNeuronsNum,
				outputNeuronsNum);
		return nnet;
	}

	/**
	 * Sets default input and output neurons for network (first layer as input,
	 * last as output)
	 */
	public static void setDefaultIO(NeuralNetwork nnet) {
		Vector<Neuron> inputNeurons = ((Layer) nnet.getLayers().firstElement())
				.getNeurons();
		Vector<Neuron> outputNeurons = ((Layer) nnet.getLayers().lastElement())
				.getNeurons();

		nnet.setInputNeurons(inputNeurons); // use first layer as input
		nnet.setOutputNeurons(outputNeurons); // use last layer as output
	}

}
